;===================================================================================================
;
;   Roger Beep Controller
;   Copyright (c) 2002-2009 OE9MKV
;
;   Port A - outputs
;   Port B - all inputs (due to pull-ups)
;
;   RA0 = PTT_out       (active high)
;   RA1 = AUD_out       (1.6 KHz)
;   RA2 = KEY_out       (active high)
;   RA3 = not used      (input)
;   RA4 = not used      (input)
;   RA5 = SPEED_LSB
;   RA6 = SPEED         (selectors for CW-speed)
;   RA7 = SPEED_MSB
;
;   RB0 = PTT_in        (active low)
;   RB1 = select "beep" (active low)
;   RB2 = select "K"    (active low)
;   RB3 = select "?"    (active low)
;   RB4 = select "R"    (active low)
;   RB5 = select "HI"   (active low)
;   RB6 = DI_inp        (active low)
;   RB7 = DAH_inp       (active low)
;
;   PTT_in LOW awakes PIC and enable PTT_out (for transceiver control)
;   PTT_in HIGH (= release) holds PTT_out and transmits Roger Beep, then releases PTT_out
;
;   AUD_out gives 1667 Hz sidetone (for "Beep", "K", "?", ...)
;
;   DI_inp LOW sends dits
;   DAH_inp LOW sends dahs
;   both inputs LOW sends alternating dits/dahs ("squeezing technique")
;
;   DI_inp LOW when PTT_in is already LOW gives 1750 Hz calling tone for European repeaters
;
;---------------------------------------------------------------------------------------------------
;
;   The original source code was written for 16F84.
;
;   Beginning with Version 2.0 the PIC is a 16F628 and uses the internal 4 MHz oscillator.
;   Disable watchdog when programming.
;
;---------------------------------------------------------------------------------------------------
;
;   Figuring out CW speed
;
;   CW speed 12 wpm is based on the word PARIS continously sent.
;   Thus PARIS is exact 5 seconds long. It needs 46 cycles.
;
;   P             A       R         I     S
;   . ___ ___ .   . ___   . ___ .   . .   . . .

;   1011101110100010111000101110100010100010101000           Mark/Space
;   12345678901234567890123456789012345678901234567890       Cycles
;
;   WPM     LPM     TFP     CYT     DIT     Bits
;                   [s]     [ms]    [ms]
;   --------------------------------------------             TFP = time for PARIS
;    12      60     5.0     109     54      000              CYT = cycle time
;    14      70     4.3      93     46      001              DIT = DI-time
;    16      80     3.8      83     41      010
;    18      90     3.3      72     36      011              WPM = words per minute
;    20     100     3.0      65     32      100              LPM = letters per minute
;    22     110     2.7      59     29      101
;    24     120     2.5      54     27      110
;    26     130     2.3      50     25      111
;
;===================================================================================================

    TITLE     "RogerBeep V2.0.628.1750"
    LIST      P=PIC16F628A
    include   "p16f628a.inc"

    __CONFIG _CP_OFF & _PWRTE_OFF & _WDT_OFF & _BODEN_OFF & _LVP_OFF &_MCLRE_OFF & _INTRC_OSC_NOCLKOUT

;===================================================================================================
;   Port Registers
;===================================================================================================

PTT_out     EQU     00H     ; PORTA / RA0
AUD_out     EQU     01H     ; PORTA / RA1
KEY_out     EQU     02H     ; PORTA / RA2

SEL_SPD0    EQU     05H     ; PORTA / RA5
SEL_SPD1    EQU     06H     ; PORTA / RA6
SEL_SPD2    EQU     07H     ; PORTA / RA7

PTT_in      EQU     00H     ; PORTB / RB0
SEL_BEEP    EQU     01H     ; PORTB / RB1
SEL_K       EQU     02H     ; PORTB / RB2
SEL_QM      EQU     03H     ; PORTB / RB3
SEL_R       EQU     04H     ; PORTB / RB4
SEL_HI      EQU     05H     ; PORTB / RB5
DI_inp      EQU     06H     ; PORTB / RB6
DAH_inp     EQU     07H     ; PORTB / RB7

;===================================================================================================
;   General Purpose Registers (GPR starts at 20H)
;===================================================================================================

A           EQU     20H     ; Accu
X           EQU     21H     ; Register X
Y           EQU     22H     ; Register Y
S           EQU     23H     ; Temporary Register S

DelayLo     EQU     24H     ; for delay()
DelayHi     EQU     25H     ; dito
Periods     EQU     26H     ; number of Periods for SOUND
CwSpeed     EQU     27H     ; selected CW speed (b'000' ... b'111')
PadFlag     EQU     28H     ; flag to decide if paddling continues (>0) or not (=0)

;===================================================================================================
;   Vectors
;===================================================================================================

        ORG    000H
        GOTO   startup

        ORG    004H
        RETURN

;===================================================================================================
;   Misc Data
;===================================================================================================

        ORG    H'008'
        DT     "RogerBeep V2.0.628.1750 * (c) OE9MKV 2002-2009."

;===================================================================================================
;   Subroutines
;===================================================================================================

        ORG    H'038'
        DT     "[ SUBS ]"

;---------------------------------------------------------------------------------------------------
;   Delay Routine for 4 MHz CPU
;   One loop takes approximately 100 us
;   Registers:  DelayLo (0..255) -> 255 * 100 us = 25.5 ms     (up to 10 ms register for decimal)
;               DelayHi (0..255) -> 255 * 25.5 ms = 6502.5 ms  (up to  1 s  register for decimal)
;               S,X,Y -> working registers (be careful - no PUSH/POP on these regs !!!)
;---------------------------------------------------------------------------------------------------

delay   MOVF    DelayHi,W       ; load parameters in working registers
        MOVWF   X               ;

dly_hi  MOVF    DelayLo,W       ;
        MOVWF   Y               ;

dly_lo  MOVLW   018H            ; [1 us]
        MOVWF   S               ; [2 us]

dly_lp  CLRWDT                  ; don't let WDT disturb us
        DECFSZ  S,F             ; inner delay loop takes 4 instruction cycles = 4 us
        GOTO    dly_lp          ; 017H * 4 us = 92 us
                                ; [93 us] after 018H cycles
        DECFSZ  Y,F             ; [94 us]
        GOTO    dly_lo          ; [96 us]
        NOP                     ; [97 us] - use 2 x NOP instead if you don't want interrupting
        NOP                     ; [98 us when this line skipped]
                                ; [104 us for only one loop will do likely]
                                ; [(DelayLo - 1) * 255 + 104 us total time for lo loop]
        DECFSZ  X,F             ;
        GOTO    dly_hi          ;

dly_xit RETURN                  ; back to who ever called us

;---------------------------------------------------------------------------------------------------
;   Generate side tone
;   This routine generates one period of the 1.67 kHz side tone signal on a 4 MHz CPU
;   1.67 kHz = 600 us period length
;---------------------------------------------------------------------------------------------------

do_tone BSF     PORTA,AUD_out
        MOVLW   01h
        MOVWF   DelayHi
        MOVLW   03h
        MOVWF   DelayLo
        CALL    delay           ; first half of wave = 300 us

        BCF     PORTA,AUD_out
        MOVLW   01h
        MOVWF   DelayHi
        MOVLW   03h
        MOVWF   DelayLo
        CALL    delay           ; second half of wave = 300 us

        RETURN

;---------------------------------------------------------------------------------------------------
;   Generate pause (1 morse code cycle)
;---------------------------------------------------------------------------------------------------

dopause MOVF    CwSpeed,W
        MOVWF   DelayHi
        MOVLW   06H             ; pause = 600 us * CwSpeed
        MOVWF   DelayLo
        CALL    delay

        RETURN

;---------------------------------------------------------------------------------------------------
;   Generate "dit" (1 morse code cycle)
;---------------------------------------------------------------------------------------------------

dodit   MOVF    CwSpeed,W
        MOVWF   Periods         ; set counter
        BSF     PORTA,KEY_out

doditlp CALL    do_tone
        DECFSZ  Periods,F       ; all periods done ?
        GOTO    doditlp          ; NO -> continue

        BCF     PORTA,KEY_out

        CALL    dopause

        RETURN

;---------------------------------------------------------------------------------------------------
;   Generate "dah" (3 morse code cycles)
;---------------------------------------------------------------------------------------------------

dodah   MOVF    CwSpeed,W
        MOVWF   Periods         ; set counter
        ADDWF   Periods,F       ; dah is 3 x dit
        ADDWF   Periods,F       ; so triple period time
        BSF     PORTA,KEY_out

dodahlp CALL    do_tone
        DECFSZ  Periods,F       ; all periods done ?
        GOTO    dodahlp         ; NO -> continue

        BCF     PORTA,KEY_out

        CALL    dopause

        RETURN

;---------------------------------------------------------------------------------------------------
;   Generate one period of the 1750 Hz audio tone
;   1750 Hz means 571 us period time, one half-wave then is 285.5 us
;   closest match is 287 us (hand-optimized) which results in 1742 Hz ...
;---------------------------------------------------------------------------------------------------

do1750  BSF     PORTA,AUD_out   ; [1 us] first half-wave starts here

        MOVLW   028H            ; [2 us] 40 * 7 us = 280 us
        MOVWF   S               ; [3 us]

fhl1750 CLRWDT                  ; [1 us] don't let WDT disturb us
        NOP                     ; [2 us]
        NOP                     ; [3 us]
        NOP                     ; [4 us]
        DECFSZ  S,F             ; [5 us] 2 us on skip, else 1 us
        GOTO    fhl1750         ; [7 us] 2 us on jump
        NOP                     ; [7 us] on exit 

        BCF     PORTA,AUD_out   ; [1 us] second half-wave starts here

        MOVLW   028H            ; [2 us] 40 * 7 us = 280 us
        MOVWF   S               ; [3 us]
        NOP                     ; [4 us]
        NOP                     ; [5 us]

shl1750 CLRWDT                  ; [1 us] don't let WDT disturb us
        NOP                     ; [2 us]
        NOP                     ; [3 us]
        NOP                     ; [4 us]
        DECFSZ  S,F             ; [5 us] 2 us on skip, else 1 us
        GOTO    shl1750         ; [7 us] 2 us on jump

        RETURN                  ; + 2 us on return gives 286 us total for second half-wave.

;---------------------------------------------------------------------------------------------------
;   Fetch delay from speed table
;---------------------------------------------------------------------------------------------------

calcspd RRF     CwSpeed,F       ; shift bits 7..5 down to 2..0
        RRF     CwSpeed,F
        RRF     CwSpeed,F
        RRF     CwSpeed,F
        RRF     CwSpeed,W
        ANDLW   B'00000111'     ; only bits 2..0 are of interest
        ADDWF   PCL,F
        RETLW   H'53'           ; "000" =  65 lpm = 50 ms = 83 peridos of 600 us
        RETLW   H'48'           ; "001" =  75 lpm = 43 ms = 72 peridos of 600 us
        RETLW   H'3F'           ; "010" =  85 lpm = 38 ms = 63 peridos of 600 us
        RETLW   H'39'           ; "011" =  95 lpm = 34 ms = 57 peridos of 600 us
        RETLW   H'34'           ; "100" = 105 lpm = 31 ms = 52 peridos of 600 us
        RETLW   H'2F'           ; "101" = 115 lpm = 28 ms = 47 peridos of 600 us
        RETLW   H'2B'           ; "110" = 125 lpm = 26 ms = 43 peridos of 600 us
        RETLW   H'28'           ; "111" = 135 lpm = 24 ms = 40 peridos of 600 us

	RETURN

;===================================================================================================
;   Main Code following
;===================================================================================================

        ORG     H'0F8'
        DT      "[ MAIN ]"

startup	BCF     STATUS,RP0      ; select BANK0
        BCF     STATUS,RP1      ; 16F628 has two bank bits

        BSF     CMCON,CM0       ; disable comparators, use RAx as I/O
        BSF     CMCON,CM1       ; 111 = all CMP inputs tied to GND
        BSF     CMCON,CM2

        BSF     STATUS,RP0      ; select BANK1
        MOVLW   B'11111000'
        MOVWF   TRISA           ;  bits [0-2] are outputs on PORTA

        MOVLW   B'11111111'
        MOVWF   TRISB           ; all bits are inputs on PORTB
        BSF     PCON,OSCF       ; select 4 MHz on internal oscillator

        MOVLW   B'00001111'     ; - enable Pulls-Ups
                                ; - enable wake-up on pin-change
                                ; - internal source for TMR0 clock
                                ; - PSA for WDT
                                ; - WDT prescaler ratio 1:128
        MOVWF   OPTION_REG      ; store selected options

wakeup  BCF     STATUS,RP0      ; select BANK0
        CLRF    PORTA		; All outputs off

	MOVLW   H'FF'           ; 2 * 25.5 ms delay because inputs may have
        MOVWF   DelayLo         ; jitter / glitches
        MOVLW   H'02'
        MOVWF   DelayHi
        CALL    delay

        MOVF    PORTA,W
        MOVWF   CwSpeed
        CALL    calcspd         ; calculate CW speed
        MOVWF   CwSpeed

        BTFSC   INTCON,RBIF     ; RB4:7 port change occured ?
        GOTO    paddlg          ; YES -> go to paddling section
        BTFSS   INTCON,INTF     ; RB0/INT occured ?
        GOTO    snooze          ; NO -> must have been something else (should never be executed !)
                                ; YES -> go to "normal" (roger-beep) section

        BTFSC   PORTB,PTT_in    ; PTT_in=0 -> skip
        GOTO    snooze          ; PTT_in=1 -> go to sleep again (must have been a glitch)

        BCF     INTCON,INTF     ; clear RB0/INT flag

ptt_on  BSF     PORTA,PTT_out   ; key transceiver up

ptt_rel CLRWDT                  ; Do not let watchdog interrupt us

chk1750 BTFSC   PORTB,DI_inp    ; PADDLE_dit pressed ? If yes, generate 1750 Hz audio tone
        GOTO    ptt_chk         ; no -> continue
        CALL    do1750          ; generate one period of 1750 Hz calling tone

ptt_chk BTFSS   PORTB,PTT_in    ; PTT_in=1 => PTT released
        GOTO    ptt_rel		; PTT_in=0 => wait for PTT_in to be released

        BTFSS   PORTB,SEL_BEEP  ; SEL_BEEP=1 -> skip
        GOTO    rgrbeep         ; SEL_BEEP=0 -> jump
        BTFSS   PORTB,SEL_K     ; SEL_K=1    -> skip
        GOTO    dadidah         ; SEL_K=0    -> jump
        BTFSS   PORTB,SEL_QM    ; SEL_QM=1   -> skip
        GOTO    qmark           ; SEL_QM=0   -> jump
        BTFSS   PORTB,SEL_R     ; SEL_R=1    -> skip
        GOTO    didadit         ; SEL_R=0    -> jump
        BTFSS   PORTB,SEL_HI    ; SEL_HI=1   -> skip
        GOTO    rgr_hi          ; SEL_HI=0   -> jump
        GOTO    final           ; no beep selected

;---------------------------------------------------------------------------------------------------

rgrbeep CALL    dodah           ; only one beep
        GOTO    final

;---------------------------------------------------------------------------------------------------

dadidah	CALL    dodah           ; -.-
        CALL    dodit
        CALL    dodah
        GOTO    final

;---------------------------------------------------------------------------------------------------

qmark   CALL    dodit            ; ..--..
        CALL    dodit
        CALL    dodah
        CALL    dodah
        CALL    dodit
        CALL    dodit
        GOTO    final

;---------------------------------------------------------------------------------------------------

didadit CALL    dodit            ; .-.
        CALL    dodah
        CALL    dodit
        GOTO    final

;---------------------------------------------------------------------------------------------------

rgr_hi  CALL    dodit            ; .... ..
        CALL    dodit
        CALL    dodit
        CALL    dodit
        CALL    dopause
        CALL    dopause
        CALL    dodit
        CALL    dodit
        GOTO    final

;---------------------------------------------------------------------------------------------------

rgr_73  CALL    dodah           ; --... ...--     not used at the moment
        CALL    dodah
        CALL    dodit
        CALL    dodit
        CALL    dodit
        CALL    dopause
        CALL    dopause
        CALL    dodit
        CALL    dodit
        CALL    dodit
        CALL    dodah
        CALL    dodah
        GOTO    final

;---------------------------------------------------------------------------------------------------

rgr_tst CALL    dodah           ; - . ... -       not used at the moment
        CALL    dopause
        CALL    dopause
        CALL    dodit
        CALL    dopause
        CALL    dopause
        CALL    dodit
        CALL    dodit
        CALL    dodit
        CALL    dopause
        CALL    dopause
        CALL    dodah
        GOTO    final

;---------------------------------------------------------------------------------------------------
;   Paddling section
;---------------------------------------------------------------------------------------------------

paddlg  BSF     PORTA,PTT_out   ; PTT_out on
        MOVLW   H'02'
        MOVWF   PadFlag         ; PadFLag := 1

pad_dit CLRWDT                  ; Do not let watchdog interrupt us
        BTFSC   PORTB,DI_inp    ; PADDLE_dit pressed ?
        GOTO    pad_dah         ; not pressed -> go on with test for "dah"
        CALL    dodit           ; send "dit"
        MOVLW   H'02'
        MOVWF   PadFlag         ; run loop at least once again

pad_dah BTFSC   PORTB,DAH_inp   ; PADDLE_dah pressed ?
        GOTO    pad_end         ; not pressed -> go on with test if paddling is finished
        CALL    dodah           ; send "dah"
        MOVLW   H'02'
        MOVWF   PadFlag         ; run loop at least once again

pad_end DECF    PadFlag         ; PadFlag := PadFlag - 1
        BTFSS   STATUS,Z        ; PadFlag = 0 ? 
        GOTO    pad_dit

;---------------------------------------------------------------------------------------------------
;   Final section ("TXTAIL") before snooze
;---------------------------------------------------------------------------------------------------

final   MOVLW   H'FF'           ; 2 * 25.5 ms delay ("TX Tail")
        MOVWF   DelayLo
        MOVLW   H'02'
        MOVWF   DelayHi
        CALL    delay

;---------------------------------------------------------------------------------------------------
;   Now falling asleep ...
;---------------------------------------------------------------------------------------------------

snooze  BCF     STATUS,RP0      ; select BANK0

        BCF	PORTA,PTT_out
        BCF	PORTA,KEY_out
        BCF	PORTA,AUD_out

        BCF     INTCON,INTF     ; clear RB0/INT flag
        BCF     INTCON,RBIF     ; clear RB4:7 port change flag

        CLRF    INTCON          ; no INTs allowed, clear flags
        BSF     INTCON,INTE     ; except RB0/INT
        BSF     INTCON,RBIE     ; and RB4:7 port change

        SLEEP                   ; fall asleep, awake when pin change occurs
        NOP                     ; for safety - don't know exactly what CPU does after wake up
        GOTO	wakeup

;===================================================================================================
;   The End
;===================================================================================================

        END
